#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

from pprint import pprint


debug_fh = None


def debug_print(*args, **kwargs):
	if debug_fh:
		print(*args, **kwargs, file=debug_fh)

def debug_pprint(*args, **kwargs):
	if debug_fh:
		pprint(*args, **kwargs, stream=debug_fh)

def set_debug_fh(fh):
	global debug_fh
	debug_fh = fh

